/*
 * Multiobjective optimization in criteria space.
 * Vladimir A. Smirnov, 2019.
 *
 * This simple program is inspired by our earlier work:
 * Гарькина, И.А. Преодоление неопределенностей целей в задаче многокрите-риальной оптимизации на примере разработки сверхтяжелых бетонов для защиты от радиации / И.А. Гарькина, А.М. Данилов, Е.В. Королев, В.А. Смирнов // Строительные материалы. — 2006. — № 8. — С. 23.
 *
 * There are some differencies, though.
 * First of all, I've used another normalization (see http://files.libv.org/contour/multiobjective.pdf).
 * The program is able to simultaneously compute several cases fo preferencies. 
 * Detailed report is printed to terminal.
 */

#include <stdio.h>
#include <math.h>

typedef	double (*funcptr_t)(double x1, double x2);

/*
 * Begin hardcoded ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

/* number of partial quality criteria, all should be maximized */
#define NUMFUNCS	6
/* number of vectors of weight coefficients */
#define NUMPRECS	6
/* centers and half-lengths of 2-dimentional factor space */
#define	X1_0		0.06
#define	DeltaX1		0.02
#define	X2_0		0.15
#define	DeltaX2		0.05

/* partial quality criteria: previously obtained regression models: */

/* thermal resistance */
double Cr(double x1,double x2)
{
	return 0.231+0.066*x1+0.017*x2-0.0071*x1*x2-0.0007*x1*x1+0.014*x2*x2;
}

/* char integrity */
double Ci(double x1,double x2)
{
	return 0.74-0.11*x1+0.13*x2-0.025*x1*x2-0.025*x1*x1-0.05*x2*x2;
}

/* durability: bases */
double Cb(double x1,double x2)
{
	return 0.79+0.015*x1-0.071*x2+0.005*x1*x2+0.036*x1*x1+0.023*x2*x2;
}

/* durability: acids */
double Ca(double x1,double x2)
{
	return 0.64-0.016*x1-0.025*x2+0.01*x1*x2+0.003*x1*x1+0.008*x2*x2;
}

/* durability: weathering */
double Cc(double x1,double x2)
{
	return 0.904-0.005*x1-0.02*x2-0.018*x1*x2+0.007*x1*x1;
}

/* durability: water */
double Cw(double x1,double x2)
{
	return 0.860+0.008*x1-0.053*x2+0.003*x1*x2+0.004*x1*x1-0.019*x2*x2;
}

funcptr_t	regmodels[NUMFUNCS] = {Cr,Ci,Cb,Ca,Cc,Cw};

/* vectors of weight coefficients */

double 		gammas[NUMPRECS][NUMFUNCS] = {
		{1,1,1,1,1,1},
		{3,3,1,1,1,1},
		{1,5,1,1,1,1},
		{1,1,3,1,1,3},
		{1,1,1,3,1,3},
		{1,1,1,1,3,3}
};

/*
 * End hardcoded ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

#define	MINMAX_DIVS	10
#define GRID_DIVS	64
double	mins[NUMFUNCS], maxs[NUMFUNCS];

__inline double to_natural(double x,double X0,double DeltaX)
{
	return x * DeltaX + X0;
}

double Theta_i(double x1,double x2,funcptr_t Q,double Qmin,double Qmax)
{
	return (Q(x1,x2)-Qmin)/(Qmax-Qmin);
}

/* FIXME */
static int global_prec_num = 0;

double Theta(double x1,double x2)
{
	int i;
	double sum = 0, theta;
	for(i = 0;i < NUMFUNCS;i++)
	{
		theta = Theta_i(x1,x2,regmodels[i],mins[i],maxs[i]) - 1;
		sum += gammas[global_prec_num][i] * theta * theta;
	}
	return sqrt(sum);
}

void
find_minmax(funcptr_t Q, /* function */
			double *min_ret, double *max_ret, /* minimum and maximum of Q */
			double *x1min_ret, double *x2min_ret, /* coordinates of minimum */
			double *x1max_ret, double *x2max_ret) /* coordinates of maximum */
{
	int i, j;
	double x1, x2, y, delta, min, max, x1min = 0, x2min = 0, x1max = 0, x2max = 0;
	delta = 2.0 / (MINMAX_DIVS - 1);
	min = max = Q(0,0);
	x2 = -1;
	for(i = 0;i < MINMAX_DIVS;i++,x2 += delta)
	{
		x1 = -1;
		for(j = 0;j < MINMAX_DIVS;j++,x1 += delta)
		{
			y = Q(x1, x2);
			if(max < y)
			{
				max = y;
				x1max = x1;
				x2max = x2;
			}
			if(min > y)
			{
				min = y;
				x1min = x1;
				x2min = x2;
			}
		}
	}
	if(min_ret)
		*min_ret = min;
	if(max_ret)
		*max_ret = max;
	if(x1min_ret)
		*x1min_ret = x1min;
	if(x2min_ret)
		*x2min_ret = x2min;
	if(x1max_ret)
		*x1max_ret = x1max;
	if(x2max_ret)
		*x2max_ret = x2max;
}

int make_grid(const char *outname,funcptr_t Q)
{
	FILE *fout;
	int i, j;
	double x1, x2, delta, zmin, zmax;
	fout = fopen(outname,"w");
	if(!fout)
	{
		printf("Unable to open file \"%s\".", outname);
		return -1;
	}
	delta = 2.0 / (GRID_DIVS - 1);
	find_minmax(Q,&zmin,&zmax,NULL,NULL,NULL,NULL);
	fprintf(fout, "DSAA\n%u %u\n",GRID_DIVS,GRID_DIVS);
	fprintf(fout, "%g %g\n", X1_0 - DeltaX1, X1_0 + DeltaX1);
	fprintf(fout, "%g %g\n", X2_0 - DeltaX2, X2_0 + DeltaX2);
	fprintf(fout, "%g %g\n", zmin, zmax);
	x2 = -1;
	for(i = 0;i < GRID_DIVS;i++,x2 += delta)
	{
		x1 = -1;
		for(j = 0;j < GRID_DIVS;j++,x1 += delta)
			fprintf(fout, "%g ", Q(x1,x2));
		fprintf(fout, "\n");
	}
	return 0;
}

int main()
{
	int i, j;
	double sum;
	char outname[256];
	double min, max, x1min, x2min, x1max, x2max;
	/* find mins and maxs for all regression models */
	for(i = 0; i < NUMFUNCS; i++)
	{
		find_minmax(regmodels[i],&min,&max,&x1min,&x2min,&x1max,&x2max);
		mins[i] = min;
		maxs[i] = max;
		printf("Model #%i: minimum %g at (%g, %g), maximum %g at (%g, %g)\n",
			i + 1,
			min, x1min, x2min,
			max, x1max, x2max);
		printf("Model #%i: minimum %g at (%g, %g), maximum %g at (%g, %g)\n",
			i + 1,
			min, to_natural(x1min,X1_0,DeltaX1), to_natural(x2min,X2_0,DeltaX2),
			max, to_natural(x1max,X1_0,DeltaX1), to_natural(x2max,X2_0,DeltaX2));
	}
	puts("");
	/* sum of gammas should be 1 */
	for(i = 0; i < NUMPRECS; i++)
	{
		sum = 0;
		for(j = 0; j < NUMFUNCS; j++)
			sum += gammas[i][j];
		sum = 1.0 / sum;
		printf("Precedency #%d: (", i);
		for(j = 0; j < NUMFUNCS; j++)
		{
			gammas[i][j] *= sum;
			printf("%g ", gammas[i][j]);
		}
		puts(")");
	}
	puts("");
#if 0
	/* grid all models */
	for(i = 0; i < NUMFUNCS; i++)
	{
		sprintf(outname,"model%02i.grd", i + 1);
		make_grid(outname, regmodels[i]);
	}
#endif
	/* find mins and maxs for every precedency */
	for(i = 0; i < NUMPRECS; i++)
	{
		global_prec_num = i;
		find_minmax(Theta,&min,&max,&x1min,&x2min,&x1max,&x2max);
		printf("Precedency #%i (code): minimum %g at (%g, %g), maximum %g at (%g, %g)\n",
			i + 1,
			min, x1min, x2min,
			max, x1max, x2max);
		printf("Precedency #%i: minimum %g at (%g, %g), maximum %g at (%g, %g)\n",
			i + 1,
			min, to_natural(x1min,X1_0,DeltaX1), to_natural(x2min,X2_0,DeltaX2),
			max, to_natural(x1max,X1_0,DeltaX1), to_natural(x2max,X2_0,DeltaX2));
		/* find values of every regression model at minimum of distance */
		puts("Values of models at optimal point: ");
		for(j = 0; j < NUMFUNCS; j++)
		{
			printf("%g\n", regmodels[j](x1min,x2min));
		}
	}
	/* grid every precedency */
	for(i = 0; i < NUMPRECS; i++)
	{
		sprintf(outname,"precedency%02i.grd", i + 1);
		global_prec_num = i;
		make_grid(outname, Theta);
	}
	return 0;
}
